;Animated LED-Heart/Arrow/Star, 18 or 20 LEDs (formerly AntiKippenLights)
;Padauk PMS150C/G or PFS154, using only PA pins (charlieplexed LEDs + button)
;For assembly with AS (asl), http://john.ccac.rwth-aachen.de:8000/as/
;
;PORTS
;  PA0   = LED matrix C
;  PA3   = LED matrix E
;  PA4   = LED matrix D
;  PA5   = button (to ground, internal pull-up)
;  PA6   = LED matrix B
;  PA7   = LED matrix A
;  PB0~7 = [unused, PFS154 only]
;
;BUTTON FUNCTION:
;  1. Press button to turn power on or off.
;  2. Hold button while connecting battery to disable auto-power-off function
;     (release button when all LEDs light up).
;  3. Hold button while connecting battery and keep holding to enter test mode
;     (for hardware debug: turn on each LED in sequence, pause and repeat).
;     Button can be released once test mode starts. Remove battery to exit.
;
;LED ARRANGEMENT
;
;   TWENTY_LEDS = 0:
;                                                15
;            12          08                        16
;          13   11    09   07           11 12 13 14  17
;         14       10       06                         18
;          15              05           10    Arrow      01
;            16   Heart  04                            02
;              17      03               09 08 07 06  03
;                18  02                            04
;                  01                            05
;
;   TWENTY_LEDS = 1:
;                                                     01
;                     17
;                       18                          02  20
;         12 13 14 15 16  19
;                           20          05   04   03      19   18   17
;         11   Long Arrow     01
;                           02              06       Star       16
;         10 09 08 07 06  03
;                       04                      07          15
;                     05                              11
;                                             08              14
;                                                  10    12
;                                           09                  13
;   PROTOTYPE_SEQUENCE = 0:
;
;                 |  ............ Anode ............
;                 |  PA3    PA4    PA0    PA6    PA7
;         --------+-----------------------------------
;          C  PA3 |  ---   LED1   LED2   LED3   LED4
;          a  PA4 | LED5    ---   LED6   LED7   LED8
;          t  PA0 | LED9   LED10   ---   LED11  LED12
;          h  PA6 | LED13  LED14  LED15   ---   LED16
;          .  PA7 | LED17  LED18 (LED19  LED20)  ---
;
;   PROTOTYPE_SEQUENCE = 1:
;
;                 |  ............ Anode ............
;                 |  PA7    PA6    PA0    PA4    PA3
;         --------+-----------------------------------
;          C  PA7 |  ---   LED1   LED2   LED3   LED4
;          a  PA6 | LED5    ---   LED6   LED7   LED8
;          t  PA0 | LED9   LED10   ---   LED11  LED12
;          h  PA4 | LED13  LED14  LED15   ---   LED16
;          .  PA3 | LED18  LED19  LED17  LED20   ---
;                   NOTE: LEDs 17/18/19 out of sequence!
;
;VERSIONS
;  v4.1  2022-01-08  Arne Rossius
;   * first version
;  v4.2  2022-02-11  Arne Rossius
;   * added different LED matrix for 18-LED projects
;   * fixed auto-power-off timeout
;   * removed erroneous extra delay in group E muxing
;  v4.3  2022-02-13  Arne Rossius
;   * added PROTOTYPE_SEQUENCE setting
;   * added auto-power-off disable functions
;   * added test mode
;  v4.4  2024-01-10  Arne Rossius
;   * changed from sdcc to AS/asl (binary output is identical to v4.3)

;===============================================================================

;SETTINGS
	IFNDEF TWENTY_LEDS
	  WARNING "TWENTY_LEDS not defined by Makefile!"

TWENTY_LEDS = 0 ;0 = 18 LEDs, 1 = 20 LEDs

	ENDIF

PROTOTYPE_SEQUENCE = 0 ;prototype arrow LED order (20 LEDs only)

;auto-power-off timeout (~14.8 kHz units, max. 2^32 - 1)
AUTO_POWER_OFF = 14815 * 60 * 60 * 2 ;approx. 2 hours
;AUTO_POWER_OFF = 14815 * 5 ;approx. 5 seconds

;button hold-down delay before power-on (1 ms units)
POWER_ON_DELAY = 50 ;50 ms

;===============================================================================

;LED SEQUENCE:
;numbers: 0 = LED1 ... 19 = LED20
;LED_x_y = LED with anode on PAx, cathode on PAy
	IF PROTOTYPE_SEQUENCE
;prototype 20-LED arrow
LED_6_7 = 0
LED_0_7 = 1
LED_4_7 = 2
LED_3_7 = 3
LED_7_6 = 4
LED_0_6 = 5
LED_4_6 = 6
LED_3_6 = 7
LED_7_0 = 8
LED_6_0 = 9
LED_4_0 = 10
LED_3_0 = 11
LED_7_4 = 12
LED_6_4 = 13
LED_0_4 = 14
LED_3_4 = 15
LED_0_3 = 16
LED_7_3 = 17
LED_6_3 = 18
LED_4_3 = 19
	ELSE
;all other boards
LED_4_3 = 0
LED_0_3 = 1
LED_6_3 = 2
LED_7_3 = 3
LED_3_4 = 4
LED_0_4 = 5
LED_6_4 = 6
LED_7_4 = 7
LED_3_0 = 8
LED_4_0 = 9
LED_6_0 = 10
LED_7_0 = 11
LED_3_6 = 12
LED_4_6 = 13
LED_0_6 = 14
LED_7_6 = 15
LED_3_7 = 16
LED_4_7 = 17
LED_0_7 = 18 ;unused if TWENTY_LEDS = 0
LED_6_7 = 19 ;unused if TWENTY_LEDS = 0
	ENDIF

	IF (TWENTY_LEDS)
N_LED = 20
	ELSE
N_LED = 18
	ENDIF
	
;===============================================================================

	SEGMENT DATA
	ORG 0x00

;Note: RAM address for t0sn/t1sn/set0/set1 must be 4 bit (0x00~0x0F)
;Note: RAM address for idxm must be 5 bit (0x00~0x1F)

	ALIGN 2
Pointer: RES 2 ;RAM pointer (16 bit, high byte must be 0)

Repeat: RES 1
Value: RES 1
Count: RES 1
PWM: RES 1
LEDs: RES 1
Button_Cnt: RES 1
Timeout: RES 4

Flags: RES 1
fAlwaysOn  BIT 0 ;always-on mode (button has no effect) - currently unused
fNoAutoOff BIT 1 ;auto-power-off disabled

Frame: RES 20

	ALIGN 2
Stack:

;===============================================================================

	SEGMENT CODE
	IF (MOMCPUNAME == "PMS150")
		INCLUDE "pms150c.inc"
		ORG FUSE_WORD
		DATA FUSE_BITS_HIGH | FUSE_SECURITY_OFF | FUSE_IO_DRV_NORMAL |\
		     FUSE_LVR_2V | FUSE_BOOTUP_SLOW
	ELSEIF (MOMCPUNAME == "PFS154")
		INCLUDE "pfs154.inc"
		ORG FUSE_WORD
		DATA FUSE_BITS_HIGH | FUSE_SECURITY_OFF | FUSE_IO_DRV_NORMAL |\
		     FUSE_BOOTUP_SLOW
	ELSE
		ERROR "CPU not supported!"
	ENDIF
	INCLUDE "easypdk.inc"

	ORG 0x000
	
	;disable Low-Voltage Reset, select fast wake-up
	mov	a, #MISC_WAKEUP_FAST_gc | 1<<MISC_LVR_DISABLE
	mov	MISC, a
	
	;set oscillator
	;switch SYSCLK to IHRC/8 (2 MHz), turn off watchdog
	mov	a, #CLKMD_IHRC_DIV8_gc | 1<<CLKMD_IHRC_EN | 1<<CLKMD_ILRC_EN
	mov	CLKMD, a
	;turn off ILRC
	mov	a, #CLKMD_IHRC_DIV8_gc | 1<<CLKMD_IHRC_EN
	mov	CLKMD, a
	
	EASYPDK_CALIBRATE_IHRC 2000000, 3000 ;2 MHz @ 3 V
	
	;init ports
	mov	a, #0x00
	mov	PA, a ;all outputs low (LEDs off)
	mov	a, #0xD9
	mov	PAC, a ;set LED pins as output
	mov	a, #0x20
	mov	PAPH, a ;enable pull-up for PA5 (button)
	mov	PADIER, a ;enable input buffer only for PA5 (button)
	IF (MOMCPUNAME == "PFS154")
	  mov	a, #0x00
	  mov	PBC, a ;set all PB pins as input
	  mov	PBDIER, a ;disable digital input buffers for all PB pins
	ENDIF
	
	;set stack pointer
	mov	a, #Stack
	mov	SP, a
	
	;init variables
	clear	Pointer+1 ;RAM pointer high byte: always 0
	clear	Flags ;TODO: find a way to set fAlwaysOn
	call	reset_timeout
	mov	a, #0
	call	set_leds ;all LEDs off
	IF (~~TWENTY_LEDS)
	  clear	Frame+18
	  clear	Frame+19
	ENDIF

	;read button (determine no auto-power-off or test mode)
	clear	Repeat
init_mode:
	;read button
	t0sn	PA, 5
	goto	init_mode_end ;button not pressed
	mov	a, #5
	call	show_frame ;5 * 1/(~235 Hz) = ~21 ms
	inc	Repeat
	mov	a, Repeat
	ceqsn	a, #47 ;~1 s
	goto	init_no_auto_off_skip
	;button held for 1 s: enable "no auto off" mode; all LEDs on
	set1	Flags, fNoAutoOff
	mov	a, #63
	call	set_leds ;all LEDs on
init_no_auto_off_skip:
	ceqsn	a, #235 ;~5 s
	goto	init_mode
	;button held for 5 s: enter test mode
	goto	test_mode
init_mode_end:
	
	;sleep
	t1sn	Flags, fAlwaysOn
	call	power_off
	;(fall through to main loop)

;-------------------------------------------------------------------------------

loop:
	;main loop
	call	animation_run_cw
	call	animation_flicker
	call	animation_worm_ccw
	call	animation_blink
	call	animation_run_ccw
	call	animation_wobble
	call	animation_worm_cw
	call	animation_glow
	goto	loop

;===============================================================================

animation_blink:
	;flash all LEDs on and off
	mov	a, #5 ;5 times
	mov	Repeat, a
	
animation_blink_loop:
	mov	a, #0x3F ;all LEDs on
	call	set_leds
	mov	a, #50
	call	show_frame
	mov	a, #0x00 ;all LEDs off
	call	set_leds
	mov	a, #50
	call	show_frame
	dzsn	Repeat
	goto	animation_blink_loop
	ret
	
;--------------------

animation_glow:
	;slowly fade all LEDs on and off
	mov	a, #3 ;3 times
	mov	Repeat, a
	mov	a, #0x00
	mov	Value, a
	
animation_glow_on:
	mov	a, Value
	call	set_leds
	mov	a, #2
	call	show_frame
	inc	Value
	t1sn	Value, 6
	goto	animation_glow_on
	
	mov	a, #0x3E
	mov	Value, a

animation_glow_off:
	mov	a, Value
	call	set_leds
	mov	a, #2
	call	show_frame
	nop
	dzsn	Value
	goto	animation_glow_off
	
	dzsn	Repeat
	goto	animation_glow_on
	ret
	
;--------------------

animation_flicker:
	;alternate between odd and even LEDs
	mov	a, #2*10 ;10 times
	mov	Repeat, a
	mov	a, #0x00 ;start with even LEDs off, odd LEDs on
	call	set_leds_odd_even
	
animation_flicker_loop:
	mov	a, #15
	call	show_frame
	call	rotate_leds_forward
	dzsn	Repeat
	goto	animation_run_cw_loop
	ret
	
;--------------------

animation_wobble:
	;slowly fade between odd and even LEDs
	mov	a, #5 ;5 times
	mov	Repeat, a
	mov	a, #0x00 ;start with even LEDs full off, odd LEDs full on
	mov	Value, a

animation_wobble_on:
	mov	a, Value
	call	set_leds_odd_even
	mov	a, #2
	call	show_frame
	inc	Value
	t1sn	Value, 6
	goto	animation_wobble_on
	
	mov	a, #0x3E
	mov	Value, a

animation_wobble_off:
	mov	a, Value
	call	set_leds_odd_even
	mov	a, #2
	call	show_frame
	nop
	dzsn	Value
	goto	animation_wobble_off
	
	dzsn	Repeat
	goto	animation_wobble_on
	ret
	
;--------------------

animation_run_cw:
	;light chaser with 2 dots, clockwise
	mov	a, #N_LED/2 * 5 ;2.5 rotations
	mov	Repeat, a
	mov	a, #0
	call	set_leds
	mov	a, #0x3F
	mov	Frame+0, a
	mov	Frame+N_LED/2, a
	
animation_run_cw_loop:
	mov	a, #20
	call	show_frame
	call	rotate_leds_forward
	dzsn	Repeat
	goto	animation_run_cw_loop
	ret
	
;--------------------

animation_run_ccw:
	;light chaser with 2 dots, counter-clockwise
	mov	a, #N_LED/2 * 5 ;2.5 rotations
	mov	Repeat, a
	mov	a, #0
	call	set_leds
	mov	a, #0x3F
	mov	Frame+N_LED/2-1, a
	mov	Frame+N_LED-1, a
	
animation_run_ccw_loop:
	mov	a, #20
	call	show_frame
	call	rotate_leds_reverse
	dzsn	Repeat
	goto	animation_run_ccw_loop
	ret
	
;--------------------

animation_worm_cw:
	;worm with fade-out tail, clockwise
	mov	a, #4 ;4 rotations
	mov	Repeat, a
	mov	a, #0x00
	call	set_leds

animation_worm_cw_start:
	mov	a, #Frame
	mov	Pointer, a

animation_worm_cw_loop:
	call	fade_leds
	mov	a, #0x3F
	idxm	Pointer, a
	inc	Pointer
	mov	a, #20
	call	show_frame
	mov	a, #Frame+N_LED
	ceqsn	a, Pointer
	goto	animation_worm_cw_loop
	
	dzsn	Repeat
	goto	animation_worm_cw_start
	ret
	
;--------------------

animation_worm_ccw:
	;worm with fade-out tail, counter-clockwise
	mov	a, #4 ;4 rotations
	mov	Repeat, a
	mov	a, #0x00
	call	set_leds

animation_worm_ccw_start:
	mov	a, #Frame+N_LED-1
	mov	Pointer, a

animation_worm_ccw_loop:
	call	fade_leds
	mov	a, #0x3F
	idxm	Pointer, a
	dec	Pointer
	mov	a, #20
	call	show_frame
	mov	a, #Frame-1
	ceqsn	a, Pointer
	goto	animation_worm_ccw_loop
	
	dzsn	Repeat
	goto	animation_worm_ccw_start
	ret
	
;===============================================================================

set_leds:
	;set all LEDs to the same brightness value
	mov	Frame+0, a
	mov	Frame+1, a
	mov	Frame+2, a
	mov	Frame+3, a
	mov	Frame+4, a
	mov	Frame+5, a
	mov	Frame+6, a
	mov	Frame+7, a
	mov	Frame+8, a
	mov	Frame+9, a
	mov	Frame+10, a
	mov	Frame+11, a
	mov	Frame+12, a
	mov	Frame+13, a
	mov	Frame+14, a
	mov	Frame+15, a
	mov	Frame+16, a
	mov	Frame+17, a
	IF (TWENTY_LEDS)
	  mov	Frame+18, a
	  mov	Frame+19, a
	ENDIF
	ret
	
;--------------------

set_leds_odd_even:
	;set all even LEDs to a brightness value,
	;set all odd LEDs to the inverse brightness value
	mov	Frame+0, a
	mov	Frame+2, a
	mov	Frame+4, a
	mov	Frame+6, a
	mov	Frame+8, a
	mov	Frame+10, a
	mov	Frame+12, a
	mov	Frame+14, a
	mov	Frame+16, a
	IF (TWENTY_LEDS)
	  mov	Frame+18, a
	ENDIF
	not	a
	and	a, #0x3F
	mov	Frame+1, a
	mov	Frame+3, a
	mov	Frame+5, a
	mov	Frame+7, a
	mov	Frame+9, a
	mov	Frame+11, a
	mov	Frame+13, a
	mov	Frame+15, a
	mov	Frame+17, a
	IF (TWENTY_LEDS)
	  mov	Frame+19, a
	ENDIF
	ret
	
;--------------------

rotate_leds_forward:
	;rotate LEDs in forward direction (LED1 -> LED2, LED2 -> LED3, ...)
	mov	a, Frame+0
	xch	Frame+1
	xch	Frame+2
	xch	Frame+3
	xch	Frame+4
	xch	Frame+5
	xch	Frame+6
	xch	Frame+7
	xch	Frame+8
	xch	Frame+9
	xch	Frame+10
	xch	Frame+11
	xch	Frame+12
	xch	Frame+13
	xch	Frame+14
	xch	Frame+15
	xch	Frame+16
	xch	Frame+17
	IF (TWENTY_LEDS)
	  xch	Frame+18
	  xch	Frame+19
	ENDIF
	mov	Frame+0, a
	ret
	
;--------------------

rotate_leds_reverse:
	;rotate LEDs in reverse direction (LED1 <- LED2, LED2 <- LED3, ...)
	IF (TWENTY_LEDS)
	  mov	a, Frame+19
	  xch	Frame+18
	  xch	Frame+17
	ELSE
	  mov	a, Frame+17
	ENDIF
	xch	Frame+16
	xch	Frame+15
	xch	Frame+14
	xch	Frame+13
	xch	Frame+12
	xch	Frame+11
	xch	Frame+10
	xch	Frame+9
	xch	Frame+8
	xch	Frame+7
	xch	Frame+6
	xch	Frame+5
	xch	Frame+4
	xch	Frame+3
	xch	Frame+2
	xch	Frame+1
	xch	Frame+0
	IF (TWENTY_LEDS)
	  mov	Frame+19, a
	ELSE
	  mov	Frame+17, a
	ENDIF
	ret
	
;--------------------

fade_leds:
	;fade each LED to half its brightness value
	sr	Frame+0
	sr	Frame+1
	sr	Frame+2
	sr	Frame+3
	sr	Frame+4
	sr	Frame+5
	sr	Frame+6
	sr	Frame+7
	sr	Frame+8
	sr	Frame+9
	sr	Frame+10
	sr	Frame+11
	sr	Frame+12
	sr	Frame+13
	sr	Frame+14
	sr	Frame+15
	sr	Frame+16
	sr	Frame+17
	IF (TWENTY_LEDS)
	  sr	Frame+18
	  sr	Frame+19
	ENDIF
	ret
	
;--------------------

show_frame:
	;show current frame 'a' times
	mov	Count, a

show_frame_loop:
	;multiplex current frame once (with PWM)
	;63 * 135 = 8505 total cycles (excluding return instruction) => ~235 Hz
	
	;running from 62 to 0: LED on if pwm < value, 100% on for value = 63
	mov	a, #62
	mov	PWM, a
pwm_loop:
	;total loop cycles:
	;	calculate A   16
	;	output A       4
	;	calculate B   17
	;	output B       6
	;	calculate C   17
	;	output C       6
	;	calculate D   17
	;	output D       6
	;	calculate E   17
	;	output E       6
	;	button etc    17
	;	output off     2
	;	loop           4
	;	================
	;	TOTAL        135
	;(2 MHz / 135) = ~14.8 kHz loop frequency
	
	;calculate data for LED group A (16 cycles)
	;PA7 = group common
	set1	LEDs, 0
	;PA6 = LED
	mov	a, PWM
	sub	a, Frame+LED_6_7
	slc	LEDs
	;PA5 = button pin
	sl	LEDs
	;PA4 = LED
	mov	a, PWM
	sub	a, Frame+LED_4_7
	slc	LEDs
	;PA3 = LED
	mov	a, PWM
	sub	a, Frame+LED_3_7
	slc	LEDs
	;PA2 & PA1 (no pin)
	sl	LEDs
	sl	LEDs
	;PA0 = LED
	mov	a, PWM
	sub	a, Frame+LED_0_7
	slc	LEDs
	
	;output data for group A (4 cycles)
	mov	a, LEDs
	mov	PAC, a ;active LEDs + group common = output (others floating)
	and	a, #~0x80 & 0xFF
	mov	PA, a ;active LEDs high, group common low
	
	;calculate data for LED group B (17 cycles)
	;PA7 = LED
	mov	a, PWM
	sub	a, Frame+LED_7_6
	slc	LEDs
	;PA6 = group common
	sl	LEDs
	set1	LEDs, 0
	;PA5 = button pin
	sl	LEDs
	;PA4 = LED
	mov	a, PWM
	sub	a, Frame+LED_4_6
	slc	LEDs
	;PA3 = LED
	mov	a, PWM
	sub	a, Frame+LED_3_6
	slc	LEDs
	;PA2 & PA1 (no pin)
	sl	LEDs
	sl	LEDs
	;PA0 = LED
	mov	a, PWM
	sub	a, Frame+LED_0_6
	slc	LEDs
	
	;output data for group B (6 cycles)
	mov	a, #0x00 ;pull previously active LED outputs low (LEDs off)
	mov	PA, a
	mov	a, LEDs
	mov	PAC, a
	and	a, #~0x40
	mov	PA, a
	
	;calculate data for LED group C (17 cycles)
	;PA7 = LED
	mov	a, PWM
	sub	a, Frame+LED_7_0
	slc	LEDs
	;PA6 = LED
	mov	a, PWM
	sub	a, Frame+LED_6_0
	slc	LEDs
	;PA5 = button pin
	sl	LEDs
	;PA4 = LED
	mov	a, PWM
	sub	a, Frame+LED_4_0
	slc	LEDs
	;PA3 = LED
	mov	a, PWM
	sub	a, Frame+LED_3_0
	slc	LEDs
	;PA2 & PA1 (no pin)
	sl	LEDs
	sl	LEDs
	;PA0 = group common
	sl	LEDs
	set1	LEDs, 0
	
	;output data for group C (6 cycles)
	mov	a, #0x00
	mov	PA, a
	mov	a, LEDs
	mov	PAC, a
	and	a, #~0x01
	mov	PA, a
	
	;calculate data for LED group D (17 cycles)
	;PA7 = LED
	mov	a, PWM
	sub	a, Frame+LED_7_4
	slc	LEDs
	;PA6 = LED
	mov	a, PWM
	sub	a, Frame+LED_6_4
	slc	LEDs
	;PA5 = button pin
	sl	LEDs
	;PA4 = group common
	sl	LEDs
	set1	LEDs, 0
	;PA3 = LED
	mov	a, PWM
	sub	a, Frame+LED_3_4
	slc	LEDs
	;PA2 & PA1 (no pin)
	sl	LEDs
	sl	LEDs
	;PA0 = LED
	mov	a, PWM
	sub	a, Frame+LED_0_4
	slc	LEDs
	
	;output data for group D (6 cycles)
	mov	a, #0x00
	mov	PA, a
	mov	a, LEDs
	mov	PAC, a
	and	a, #~0x10
	mov	PA, a
	
	;calculate data for LED group E (17 cycles)
	;PA7 = LED
	mov	a, PWM
	sub	a, Frame+LED_7_3
	slc	LEDs
	;PA6 = LED
	mov	a, PWM
	sub	a, Frame+LED_6_3
	slc	LEDs
	;PA5 = button pin
	sl	LEDs
	;PA4 = LED
	mov	a, PWM
	sub	a, Frame+LED_4_3
	slc	LEDs
	;PA3 = group common
	sl	LEDs
	set1	LEDs, 0
	;PA2 & PA1 (no pin)
	sl	LEDs
	sl	LEDs
	;PA0 = LED
	mov	a, PWM
	sub	a, Frame+LED_0_3
	slc	LEDs
	
	;output data for group E (6 cycles)
	mov	a, #0x00
	mov	PA, a
	mov	a, LEDs
	mov	PAC, a
	and	a, #~0x08
	mov	PA, a
	
	;poll push-button (8 cycles unless power_off is called)
	t0sn	PA, 5
	goto	button_released
	;button pressed
	t1sn	Button_Cnt, 7 ;prevent overflow
	inc	Button_Cnt
	mov	a, #127
	ceqsn	a, Button_Cnt ;pressed for (127 / 8.4 kHz) = ~15 ms ?
	goto	button_end
	t1sn	Flags, fAlwaysOn
	call	power_off
	goto	button_end
button_released:
	clear	Button_Cnt
	goto	*+1 ;2 cycles delay
	goto	*+1 ;2 cycles delay
button_end:

	;auto-power-off timeout counter (8 cycles unless power_off is called)
	add	a, #0 ;clear carry flag
	t1sn	Flags, fNoAutoOff
	dec	Timeout+0
	subc	Timeout+1
	subc	Timeout+2
	subc	Timeout+3
	t0sn	FLAG, FLAG_C
	call	power_off
	
	;extra delay: button (8) + timeout (8) + 1 = 17 cycles
	nop ;1 cycle
	
	;output off (2 cycles)
	mov	a, #0x00 ;pull previously active LED outputs low (LEDs off)
	mov	PA, a
	
	;end of PWM loop (4 cycles when looping, 3+ret when exiting loop)
	dec	PWM
	t1sn	FLAG, FLAG_C
	goto	pwm_loop
	
	;end of show frame loop
	dzsn	Count
	goto	show_frame_loop
	ret

;--------------------

power_off:
	;button pressed or auto-power-off timeout elapsed: turn off
	
	;turn off LED group C
	mov	a, #0x00 ;pull previously active LED outputs low (LEDs off)
	mov	PA, a
	mov	a, #0xD9 ;set all LED pins as outputs
	mov	PAC, a
	
	;wait until button released
power_off_wait:
	t1sn	PA, 5
	goto	power_off_wait
	
power_off_sleep:
	;sleep (wake up when any pin enabled in PADIER/PBDIER toggles)
	stopsys
	
	;check if button pressed for a certain time (filter ESD and spikes)
	clear	Button_Cnt
power_on_loop:
	;delay ~1 ms (~2000 cycles)
	mov	a, #3
	mov	PWM, a
	mov	a, #153
power_on_delay_loop:
	dzsn	a
	goto	power_on_delay_loop
	dzsn	PWM
	goto	power_on_delay_loop
	;poll button
	t0sn	PA, 5
	goto	power_off_sleep
	;button pressed
	inc	Button_Cnt
	mov	a, #POWER_ON_DELAY
	ceqsn	a, Button_Cnt
	goto	power_on_loop

	;power on
	mov	a, #0xFF
	mov	Button_Cnt, a ;avoid duplicate button press detection
	
	;reset power-off timeout
reset_timeout:
	mov	a, #AUTO_POWER_OFF & 0xFF
	mov	Timeout+0, a
	mov	a, #AUTO_POWER_OFF>>8 & 0xFF
	mov	Timeout+1, a
	mov	a, #AUTO_POWER_OFF>>16 & 0xFF
	mov	Timeout+2, a
	mov	a, #AUTO_POWER_OFF>>24
	mov	Timeout+3, a
	
	ret

;===============================================================================

test_led MACRO ANODE, CATHODE
	mov	a, #(1<<ANODE | 1<<CATHODE) ;set anode & cathode pins to output
	mov	PAC, a
	mov	a, #(1<<ANODE) ;set anode pin high
	mov	PA, a
	call	test_delay
	ENDM

test_mode:
	;test mode for hardware debugging
	
	;turn on LEDs one at a time in sequence
	;(sequence is hard-coded for PROTOTYPE_SEQUENCE = 0)
	test_led 4, 3 ;LED 1: PA4 ---|>|--- PA3
	test_led 0, 3 ;LED 2
	test_led 6, 3 ;LED 3
	test_led 7, 3 ;LED 4
	test_led 3, 4 ;LED 5
	test_led 0, 4 ;LED 6
	test_led 6, 4 ;LED 7
	test_led 7, 4 ;LED 8
	test_led 3, 0 ;LED 9
	test_led 4, 0 ;LED 10
	test_led 6, 0 ;LED 11
	test_led 7, 0 ;LED 12
	test_led 3, 6 ;LED 13
	test_led 4, 6 ;LED 14
	test_led 0, 6 ;LED 15
	test_led 7, 6 ;LED 16
	test_led 3, 7 ;LED 17
	test_led 4, 7 ;LED 18
	IF (TWENTY_LEDS)
	  test_led 0, 7 ;LED 19
	  test_led 6, 7 ;LED 20
	ENDIF
	
	;all LEDs off
	mov	a, #0
	mov	PAC, a
	mov	PA, a
	call	test_delay
	
	;repeat indefinitely
	goto	test_mode
	
;--------------------

test_delay:
	;delay: 5 * (2 + 256 * (2 + 256 * 3)) = 985610 cycles = ~493 ms
	mov	a, #5
	mov	Count, a
	clear	PWM
	mov	a, #0
test_delay_loop:
	dzsn	a
	goto	test_delay_loop
	dzsn	PWM
	goto	test_delay_loop
	dzsn	Count
	goto	test_delay_loop
	ret

;===============================================================================

	IF (MOMPASS == 2)
	  OUTRADIX 10
	  SEGMENT CODE
	  MESSAGE "ROM use: \{*} words."
	  SEGMENT DATA
	  MESSAGE "RAM use: \{*} bytes."
	  MESSAGE "Auto Power Off: \{AUTO_POWER_OFF/14815/60} minutes."
	  IF (TWENTY_LEDS)
	    MESSAGE "*** Assembling for 20 LEDs ***"
	  ELSE
	    MESSAGE "*** Assembling for 18 LEDs ***"
	  ENDIF
	  IF (PROTOTYPE_SEQUENCE)
	    WARNING "##### Prototype sequence selected! #####"
	  ENDIF
	ENDIF
